# coding=utf-8
from __future__ import print_function

import pandas as pd
from pyhive import hive

from suanpan.log import logger


class HiveDataWarehouse(object):

    DEFAULT_HIVE_DB_NAME = "default"
    DTYPE_SQLTYPE_MAPPING = {
        "int8": "tinyint",
        "int16": "smallint",
        "int32": "int",
        "int64": "bigint",
        "float32": "float",
        "float64": "double",
        "class": "string",
        "datetime": "timestamp",
    }

    def __init__(
        self,  # pylint: disable=unused-argument
        hiveHost,
        hivePort=10000,
        hiveDatabase="default",
        hiveUsername=None,
        hivePassword=None,
        hiveAuth=None,
        **kwargs
    ):
        self.host = hiveHost
        self.port = hivePort
        self.database = hiveDatabase
        self.username = hiveUsername
        self.password = hivePassword
        self.auth = hiveAuth

        self._connection = None

    def connect(self):
        connection = hive.Connection(
            host=self.host,
            port=self.port,
            database=self.database,
            username=self.username,
            password=self.password,
            auth=self.auth,
        )
        return connection

    def testConnection(self, connection):
        with connection.cursor() as cursor:
            self.selectCurrentTimestamp(cursor)
        logger.info("Hive connected.")

    def readTable(self, table, partition=None, database=DEFAULT_HIVE_DB_NAME):
        sql = u"select * from {}.{}".format(database, table)

        if partition and partition.strip():
            conditions = u" and ".join(x.strip() for x in partition.split(","))
            sql = u"{} where {}".format(sql, conditions)

        with self.connect() as connection:
            data = pd.read_sql(sql, connection)
        data = data.rename(index=str, columns=lambda x: x.split(".")[-1])
        return data

    def writeTable(self, table, data, database=DEFAULT_HIVE_DB_NAME, overwrite=True):
        columns = self.getColumns(data)
        with self.connect() as connection:
            with connection.cursor() as cursor:
                self.useDatabase(cursor, database)
                if overwrite:
                    self.dropTable(cursor, table)
                self.createTable(cursor, table, columns)
                self.insertData(cursor, table, data)

    def execute(self, sql):
        with self.connect() as connection:
            with connection.cursor() as cursor:
                return self._execute(cursor, sql)

    def _execute(self, cursor, sql):
        try:
            return cursor.execute(sql)
        except Exception as e:
            logger.error("SQL: {}".format(sql))
            raise e

    def createTable(self, cursor, table, columns):
        self._execute(
            cursor,
            u"create table if not exists {table} ({columns})".format(
                table=table,
                columns=u", ".join(
                    [u"{} {}".format(c["name"], c["type"]) for c in columns]
                ),
            ),
        )

    def dropTable(self, cursor, table):
        self._execute(cursor, u"drop table {}".format(table))

    def insertData(self, cursor, table, data, mode="into", per=10000):
        dataLength = len(data)
        logger.info(u"Insert into tabel {}: total {}".format(table, dataLength))
        pieces = dataLength // per + 1
        for i in range(pieces):
            begin = per * i
            end = per * (i + 1)
            values = data[begin:end]
            self.insertPiece(cursor, table, values, mode=mode, begin=begin)

    def insertPiece(self, cursor, table, data, mode="into", begin=0):
        if not data.empty:
            end = begin + len(data)
            logger.info(u"[{} - {}] inserting...".format(begin, end))
            valueGroup = lambda d: u"({})".format(
                u", ".join([u'"{}"'.format(i) for i in d])
            )
            self._execute(
                cursor,
                u"insert {mode} {table} values {values}".format(
                    table=table,
                    values=u", ".join(data.apply(valueGroup, axis=1).values),
                    mode=mode,
                ),
            )
            logger.info(u"[{} - {}] Done.".format(begin, end))

    def useDatabase(self, cursor, database):
        self._execute(cursor, u"use {}".format(database))

    def selectCurrentTimestamp(self, cursor):
        self._execute(cursor, u"select current_timestamp()")

    def getColumns(self, data):
        return [
            {"name": name.split(".")[-1], "type": self.toSqlType(dtype)}
            for name, dtype in data.dtypes.items()
        ]

    def toSqlType(self, dtype):
        typeName = self.shortenDatetimeType(dtype.name)
        return self.DTYPE_SQLTYPE_MAPPING.get(typeName, "string")

    def shortenDatetimeType(self, typeName):
        datatimeType = "datetime"
        return datatimeType if typeName.startswith(datatimeType) else typeName



# coding=utf-8
from __future__ import print_function

import itertools
import tempfile

from suanpan.arguments import Bool, Int, String
from suanpan.proxy import Proxy
from suanpan.storage.local import LocalStorage
from suanpan.storage.minio import MinioStorage
from suanpan.storage.oss import OssStorage

DEFAULT_TEMP_STORE = tempfile.gettempdir()


class StorageProxy(Proxy):
    MAPPING = {"oss": OssStorage, "local": LocalStorage, "minio": MinioStorage}
    DEFAULT_ARGUMENTS = [String("storage-type")]
    OSS_ARGUMENTS = [
        String("storage-oss-access-id"),
        String("storage-oss-access-key"),
        String("storage-oss-bucket-name", default="suanpan"),
        String("storage-oss-endpoint", default="http://oss-cn-beijing.aliyuncs.com"),
        String("storage-oss-delimiter", default="/"),
        String("storage-oss-temp-store", default=DEFAULT_TEMP_STORE),
        Int("storage-oss-download-num-threads", default=1),
        String("storage-oss-download-store-name", default=".py-oss-download"),
        Int("storage-oss-upload-num-threads", default=1),
        String("storage-oss-upload-store-name", default=".py-oss-upload"),
    ]
    LOCAL_ARGUMENTS = [String("storage-local-temp-store", default=DEFAULT_TEMP_STORE)]
    MINIO_ARGUMENTS = [
        String("storage-minio-access-key"),
        String("storage-minio-secret-key"),
        String("storage-minio-bucket-name", default="suanpan"),
        String("storage-minio-endpoint"),
        Bool("storage-minio-secure", default=True),
        String("storage-minio-delimiter", default="/"),
        String("storage-minio-temp-store", default=DEFAULT_TEMP_STORE),
    ]
    ARGUMENTS = list(
        itertools.chain(
            DEFAULT_ARGUMENTS, OSS_ARGUMENTS, LOCAL_ARGUMENTS, MINIO_ARGUMENTS
        )
    )


storage = StorageProxy()



# coding=utf-8
from __future__ import print_function

import functools
import traceback

import retrying

from suanpan.log import logger


def _log(func):
    @functools.wraps(func)
    def _dec(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logger.warning("Run failed and retrying: {}".format(func.__name__))
            logger.warning(traceback.format_exc())
            raise e

    return _dec


def retry(*args, **kwargs):
    def _wrap(func):
        kwargs.update(wrap_exception=True)
        _retry = retrying.retry(*args, **kwargs)
        _func = _retry(_log(func))

        @functools.wraps(func)
        def _dec(*fargs, **fkwargs):
            try:
                return _func(*fargs, **fkwargs)
            except retrying.RetryError as e:
                logger.error(
                    "Retry failed after {} attempts: {}".format(
                        e.last_attempt.attempt_number, func.__name__
                    )
                )
                raise e

        return _dec

    return _wrap



# coding=utf-8
from __future__ import print_function

from addict import Dict


class HasName(object):
    @property
    def name(self):
        return self.__class__.__name__


class Context(Dict, HasName):
    pass


class Global(Context):
    pass
